==> de.flapdoodle.reverse-1.5.0-sources.jar: de/flapdoodle/reverse/Transitions.java <==
/**
 * Copyright (C) 2016
 *   Michael Mosmann <michael@mosmann.de>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *         http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package de.flapdoodle.reverse;

import de.flapdoodle.checks.Preconditions;
import de.flapdoodle.graph.GraphAsDot;
import de.flapdoodle.graph.Graphs;
import de.flapdoodle.graph.ImmutableSubGraph;
import de.flapdoodle.reverse.naming.HasLabel;
import de.flapdoodle.reverse.types.TypeNames;
import de.flapdoodle.types.Either;
import org.immutables.value.Value;
import org.jgrapht.graph.DefaultDirectedGraph;
import org.jgrapht.graph.DefaultEdge;

import java.util.*;
import java.util.function.Function;
import java.util.function.Supplier;
import java.util.stream.Collectors;

@Value.Immutable
public abstract class Transitions {

	public abstract List<Transition<?>> transitions();

	@Value.Auxiliary
	public Transitions addAll(Transition<?> ...transitions) {
		return ImmutableTransitions.builder()
			.from(this)
			.addTransitions(transitions)
			.build();
	}

	public Transitions addAll(Transitions other) {
		return ImmutableTransitions.builder()
			.from(this)
			.addAllTransitions(other.transitions())
			.build();
	}

	@Value.Auxiliary
	public <T> Transitions replace(Transition<T> transition) {
		List<Transition<?>> filteredTransition = transitions().stream()
			.filter(it -> !it.destination().equals(transition.destination()))
			.collect(Collectors.toList());

		boolean transitionWithSameDestinationRemoved = filteredTransition.size() + 1 == transitions().size();

		Preconditions.checkArgument(transitionWithSameDestinationRemoved,"no transition with destination %s found", transition.destination());

		return ImmutableTransitions.builder()
			.transitions(filteredTransition)
			.addTransitions(transition)
			.build();
	}

	@Value.Check
	protected void checkForCollisions() {
		assertNoCollisions(transitions());
	}

	@Value.Auxiliary
	public DefaultDirectedGraph<Vertex, DefaultEdge> asGraph() {
		return asGraph(transitions());
	}

	@Value.Auxiliary
	public TransitionWalker walker() {
		return TransitionWalker.with(transitions());
	}

	public static Transitions from(Transition<?> ... transitions){
		return ImmutableTransitions.builder().addTransitions(transitions).build();
	}

	public static DefaultDirectedGraph<Vertex, DefaultEdge> asGraph(List<? extends Transition<?>> all) {
		Supplier<Graphs.GraphBuilder<Vertex, DefaultEdge, DefaultDirectedGraph<Vertex, DefaultEdge>>> directedGraph = Graphs
			.graphBuilder(Graphs.directedGraph(DefaultEdge.class));
		return Graphs.with(directedGraph).build(graph -> all.forEach(edge -> {
			StateVertex destination = StateVertex.of(edge.destination());
			TransitionVertex transition = TransitionVertex.of(edge);

			graph.addVertex(destination);
			graph.addVertex(transition);
			graph.addEdge(transition, destination);
			edge.sources().forEach(source -> {
				StateVertex s = StateVertex.of(source);
				graph.addVertex(s);
				graph.addEdge(s, transition);
			});
		}));
	}

	public static void assertNoCollisions(List<? extends Transition<?>> all) {
		String transitionWithCollisions = all.stream()
			.collect(Collectors.groupingBy(Transition::destination))
			.entrySet().stream().filter(entry -> entry.getValue().size() > 1)
			.map(entry -> entry.getKey() + " --> "+entry.getValue())
			.collect(Collectors.joining(",\n  "));

		Preconditions.checkArgument(transitionWithCollisions.isEmpty(), "multiple transitions with same destination: \n  %s", transitionWithCollisions);
	}

	public static String edgeGraphAsDot(String label, DefaultDirectedGraph<Vertex, DefaultEdge> graph) {
		return edgeGraphAsDot(label, graph, Transitions::transitionAsLabel, Transitions::stateAsLabel);
	}

	public static String edgeGraphAsDot(String label, DefaultDirectedGraph<Vertex, DefaultEdge> graph,
		Function<Transition<?>, String> transitionAsLabel,
		Function<StateID<?>, String> stateIdAsLabel
	) {
		Function<Vertex, String> vertexAsLabel = t -> {
			Either<StateVertex, TransitionVertex> stateOrTransition = asEither(t);

			return stateOrTransition
				.mapLeft(StateVertex::stateId)
				.mapLeft(stateIdAsLabel::apply)
				.mapRight(TransitionVertex::transition)
				.mapRight(transitionAsLabel::apply)
				.map(Function.identity(), Function.identity());
		};

		return GraphAsDot.builder(Transitions.asId())
			.subGraphIdSeparator("__")
			.label(label)
			.nodeAsLabel(vertexAsLabel)
			.sortedBy((GraphAsDot.AsComparable<Vertex, String>) vertexAsLabel::apply)
			.nodeAttributes(t -> {
				Either<StateVertex, TransitionVertex> stateOrTransition = asEither(t);
				String shape = stateOrTransition.isLeft() ? "ellipse" : "rectangle";
				return asMap("shape", shape);
			})
			.subGraph(t -> {
				if (t instanceof TransitionVertex) {
					Transition<?> transition = ((TransitionVertex) t).transition();
					if (transition instanceof TransitionWalker.MappedWrapper) {
						TransitionWalker.MappedWrapper<?> wrapper = (TransitionWalker.MappedWrapper<?>) transition;
						ImmutableSubGraph<Vertex> subGraph = GraphAsDot.SubGraph.of(wrapper.graph())
							.connections(asSubGraphMap(wrapper.transitionMapping(), wrapper.missingSources()))
							.build();

						return Optional.of(subGraph);
					}
				}
				return Optional.empty();
			})
			.build().asDot(graph);
	}

	private static Map<? extends Vertex, ? extends Vertex> asSubGraphMap(TransitionMapping<?> transitionMapping, Set<StateID<?>> missingSources) {
		Map<Vertex, Vertex> ret=new LinkedHashMap<>();
		missingSources.forEach(dest -> {
			StateID<?> source = transitionMapping.sourceOf(dest);
			ret.put(StateVertex.of(source), StateVertex.of(dest));
		});
		ret.put(StateVertex.of(transitionMapping.destination().destination()), StateVertex.of(transitionMapping.destination().source()));

		return Collections.unmodifiableMap(ret);
	}

	private static String stateAsLabel(StateID<?> t) {
		return (t.name().isEmpty() ? "<empty>" : t.name()) + ":" + TypeNames.typeName(t.type());
	}

	private static String transitionAsLabel(Transition<?> route) {
		if (route instanceof HasLabel) return ((HasLabel) route).transitionLabel();
		return TypeNames.typeName(route.getClass());
	}

	private static Map<String, String> asMap(String... keyValues) {
		LinkedHashMap<String, String> ret = new LinkedHashMap<>();
		if ((keyValues.length % 2) != 0) {
			throw new IllegalArgumentException("parameter not modulo of 2");
		}
		for (int i = 0; i < keyValues.length; i = i + 2) {
			ret.put(keyValues[i], keyValues[i + 1]);
		}
		return ret;
	}

	private static Function<Vertex, String> asId() {
		LinkedHashMap<Class<?>, Integer> typeCounter=new LinkedHashMap<>();
		LinkedHashMap<Transition<?>, Integer> transitionMap = new LinkedHashMap<>();

		return vertex -> asEither(vertex)
			.mapLeft(StateVertex::stateId)
			.mapLeft(type -> (type.name().isEmpty() ? "<empty>" : type.name()) + ":" + type.type().toString())
			.mapRight(TransitionVertex::transition)
			.mapRight(transition -> {
				Integer number= transitionMap.get(transition);
				if (number==null) {
					typeCounter.putIfAbsent(transition.getClass(),-1);
					number = typeCounter.get(transition.getClass()) + 1;
					typeCounter.put(transition.getClass(), number);
					transitionMap.put(transition, number);
				}
				return transition.getClass().getName() +":"+ number;
			})
			.map(Function.identity(), Function.identity());
	}

	public static Either<StateVertex, TransitionVertex> asEither(Vertex vertex) {
		if (vertex instanceof StateVertex) return Either.left((StateVertex) vertex);
		if (vertex instanceof TransitionVertex) return Either.right((TransitionVertex) vertex);
		throw new IllegalArgumentException("unknown vertex type: " + vertex + "(" + vertex.getClass() + ")");
	}

	interface Vertex {

	}

	@Value.Immutable
	interface StateVertex extends Vertex {
		@Value.Parameter
		StateID<?> stateId();

		static StateVertex of(StateID<?> stateId) {
			return ImmutableStateVertex.of(stateId);
		}
	}

	@Value.Immutable
	interface TransitionVertex extends Vertex {
		@Value.Parameter
		Transition<?> transition();

		static TransitionVertex of(Transition<?> transition) {
			return ImmutableTransitionVertex.of(transition);
		}
	}
}

==> de.flapdoodle.reverse-1.7.2-sources.jar: de/flapdoodle/reverse/Transitions.java <==
/*
 * Copyright (C) 2016
 *   Michael Mosmann <michael@mosmann.de>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *         http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package de.flapdoodle.reverse;

import de.flapdoodle.checks.Preconditions;
import de.flapdoodle.reverse.graph.*;
import org.immutables.value.Value;
import org.jgrapht.graph.DefaultDirectedGraph;
import org.jgrapht.graph.DefaultEdge;

import java.util.*;
import java.util.stream.Collectors;

@Value.Immutable
public abstract class Transitions {

	public abstract List<Transition<?>> transitions();

	@Value.Auxiliary
	public Transitions addAll(Transition<?> ...transitions) {
		return ImmutableTransitions.builder()
			.from(this)
			.addTransitions(transitions)
			.build();
	}

	public Transitions addAll(Transitions other) {
		return ImmutableTransitions.builder()
			.from(this)
			.addAllTransitions(other.transitions())
			.build();
	}

	@Value.Auxiliary
	public <T> Transitions replace(Transition<T> transition) {
		List<Transition<?>> filteredTransition = transitions().stream()
			.filter(it -> !it.destination().equals(transition.destination()))
			.collect(Collectors.toList());

		boolean transitionWithSameDestinationRemoved = filteredTransition.size() + 1 == transitions().size();

		Preconditions.checkArgument(transitionWithSameDestinationRemoved,"no transition with destination %s found", transition.destination());

		return ImmutableTransitions.builder()
			.transitions(filteredTransition)
			.addTransitions(transition)
			.build();
	}

	@Value.Check
	protected void checkForCollisions() {
		assertNoCollisions(transitions());
	}

	@Value.Auxiliary
	public TransitionWalker walker() {
		return TransitionWalker.with(transitions());
	}

	public static Transitions from(Transition<?> ... transitions){
		return ImmutableTransitions.builder().addTransitions(transitions).build();
	}

	public static void assertNoCollisions(List<? extends Transition<?>> all) {
		String transitionWithCollisions = all.stream()
			.collect(Collectors.groupingBy(Transition::destination))
			.entrySet().stream().filter(entry -> entry.getValue().size() > 1)
			.map(entry -> entry.getKey() + " --> "+entry.getValue())
			.collect(Collectors.joining(",\n  "));

		Preconditions.checkArgument(transitionWithCollisions.isEmpty(), "multiple transitions with same destination: \n  %s", transitionWithCollisions);
	}
}
